/* Una matriu en pike és un objecte de tipus immutable. La forma de simular objectes mutables és convertir-la en un array i treballar amb ell. Després simplement es crea una nova matriu a partir de l'array */

//! A NMatrix object is a matrix.
class NMatrix {
  
  inherit Math.Matrix ;
  
    //! You can create a NMatrix object giving an array or two ints: the number of rows and columns.
  void create( int|array fst , void|int snd )
  {
    if(snd) //Si rep dos arguments, és que cal crear una matriu a partir de les dimensions. Gire els índex per a evitar problemes amb la transposta
      { 
	::create(snd,fst);
      }
    else //Si rep un sol argument, cride al constructor de la classe pare per a crear una matriu a partir d'un array.
      {
	::create(fst);
      }
  }
  
  //! Using get you obtain the element at row i, column j of the NMatrix.
  float get(  int i , int j ) 
  {
    array mat = cast("array");
    return  mat[i][j] ;
  }
  
  //!Returns an array ({ rows, columns })
  public array(int) dims()
  {
    int files, columnes ;
    array mat = cast("array");
    files = sizeof(mat);
    columnes = sizeof(mat[0] ) ;
    return ( { files, columnes } ) ;
  }
  
  //! Number of rows of a NMatrix
  public int rows ()
  {
    array mat = cast("array") ;
    return sizeof(mat);
  }
  
  //!Number of columns of a NMatrix
  public int columns()
  {
    array mat = cast("array");
    return sizeof(mat[0]);
  }
  
  //!Shows a NMatrix on the screen.
  public void show () 
  {
    int files = dims()[0] ;
    int columnes = dims()[1] ;
    array(array(float|int) ) ar = cast("array" ) ;
    for (int i = 0 ; i < files ; ++i )  
      {
	write( "[ " ) ;
	for ( int j = 0 ; j < columnes ; ++j )
	  {
	    write("" + ar[i][j]) ;
	    if (j == (columnes - 1 ) )
	      {
		write( " " ) ;
	      }
	    else 
	      {
		write( " , " ) ;
	      }
	  }
	write( " ] \n" );
      }
  }
  
  //!Multiplies a NMatrix by a scalar, multiplying all its elements by that scalar.
  public NMatrix productByScalar ( int|float n ) 
  {
    NMatrix resultat ;
    array(array(int|float) ) mat = cast("array" ) ;
    int files = dims()[0] ;
    int columnes = dims()[1] ;
    for( int i = 0 ; i < files ; ++i )
      {
	for (int j = 0 ; j < columnes ; ++j ) 
	  {
	    mat[i][j] = mat[i][j] * n ;
	  }
      }
    resultat = NMatrix(mat);
    return resultat ;
  }
  
  //!Returns a new NMatrix, with rows i and j swapped.
  public NMatrix swapRows ( int i,int j )
  {
    int files = rows();
    int columnes = columns();
    array(array(float|int) ) mat = cast("array") ;
    float|int auxiliar ;
    NMatrix resultat ;
    for ( int k = 0 ; k < columnes ; ++k )
      {
	auxiliar = mat[i][k]  ;
	mat[i][k] = mat[j][k] ;
	mat[j][k] = auxiliar ;
      }
    resultat = NMatrix(mat);
    return resultat ;
  }
  
  //!Multiplies two matrices.
  public NMatrix mult ( NMatrix b ) 
  {
    int files1 = rows() ;
    int files2 = b->rows();
    int columnes1 = columns();
    int columnes2 = b->columns();
    if ( columnes1 == files2  )
      {
	int files = files1 ; 
	int columnes = columnes2 ;
	//write("Multiplicant " + files1 + " x " + columnes1 + " per " + files2 + " x " + columnes2 + " per donar " + files + " x " + columnes + "\n");
	int aux = columns();
	array a1 = cast("array");
	array a2 = b->cast("array") ;
	NMatrix aux2  = NMatrix(  files, columnes ); 
	array resultat = aux2->cast("array");
	float suma = 0 ;
	//	write("Resultat té dimensió " + sizeof(resultat) + " x " + sizeof(resultat[0] ) + "\n") ;
	for ( int i = 0; i < files ; ++i)
	  {
	    for ( int j = 0 ; j < columnes ; ++j )
	      {
		suma = 0 ;
		for ( int k = 0 ; k < aux ; ++k )
		  {
		    suma += a1[i][k] * a2[k][j] ;
		  }
		resultat[i][j] = suma ;
	      }
	  }
	return NMatrix(resultat);
      }
    else
      {
	return 0;
      }
  }
  
  //! a * b 
  public NMatrix `* (NMatrix a )
  {
   return  mult(a) ;
  }

  //! a - b 
  public NMatrix sub( NMatrix b ) 
  {
    int files1 = rows() ;
    int files2 = b->rows();
    int columnes1 = columns();
    int columnes2 = b->columns();
    if ( files1 == files2  && columnes1 == columnes2  )
      {
	
	array a1 = cast("array");
	array a2 = b->cast("array") ;
	NMatrix aux2  = NMatrix(  files1, columnes1 ); 
	array resultat = aux2->cast("array");
	float suma = 0 ;
	//	write("Resultat té dimensió " + sizeof(resultat) + " x " + sizeof(resultat[0] ) + "\n") ;
	for ( int i = 0; i < files1 ; ++i)
	  {
	    for ( int j = 0 ; j < columnes1 ; ++j )
	      {
		
		resultat[i][j] = a1[i][j] - a2[i][j] ;
	      }
	  }
	return NMatrix(resultat);
      }
    else
      {
	return 0;
      }
  }
  
  //! a + b 
   public NMatrix sum( NMatrix b ) 
  {
    int files1 = rows() ;
    int files2 = b->rows();
    int columnes1 = columns();
    int columnes2 = b->columns();
    if ( files1 == files2  && columnes1 == columnes2  )
      {
	
	array a1 = cast("array");
	array a2 = b->cast("array") ;
	NMatrix aux2  = NMatrix(  files1, columnes1 ); 
	array resultat = aux2->cast("array");
	float suma = 0 ;
	//	write("Resultat té dimensió " + sizeof(resultat) + " x " + sizeof(resultat[0] ) + "\n") ;
	for ( int i = 0; i < files1 ; ++i)
	  {
	    for ( int j = 0 ; j < columnes1 ; ++j )
	      {
		
		resultat[i][j] = a1[i][j] + a2[i][j] ;
	      }
	  }
	return NMatrix(resultat);
      }
    else
      {
	return 0;
      }
  }
  
  //! a + b
  public NMatrix `+(NMatrix b ) 
  {
    return  sum(b);
  }
  
  //! a - b
  public NMatrix `-(NMatrix b )
  {
    return sub(b);
  }
  
  //! Returns a new matrix, transposed of the original matrix.
  public NMatrix transpose ( )
  {
    array mat = cast("array");
    int files = rows();
    int columnes = columns();
    array resultat = NMatrix( columnes, files)->cast("array");
    for ( int i = 0 ; i < files ; ++i )
      {
	for ( int j = 0 ; j < columnes ; ++j )
	  {
	    resultat[j][i] = mat[i][j] ;
	  }
      }
    return NMatrix(resultat);
  }
  
  //!Returns a NMatrix object, with the same content of the i column of the original matrix.
  public NMatrix getColumn(int i )
  {
    int files = rows();
    NMatrix aux = NMatrix(files, 1);
    array resultat = aux->cast("array");
    array mat = cast("array");
    for(int j = 0 ; j < files ; ++j )
      {
	resultat[j][0] = mat[j][i] ;
      }
    return NMatrix(resultat);
  }
  
  //!This does the LU decomposition of the matrix. Returns an array ({ L , U , permutation_matrix , signum })
  public array LU () 
  {
    array(array(float|int)) copia = cast("array");
    int dimensio = rows();
    array l = NMatrix(dimensio, dimensio)->cast("array") ;
    array u = NMatrix(dimensio, dimensio)->cast("array") ;
    NMatrix permutacions = NMatrix(dimensio, dimensio) ;
    NMatrix auxiliar ;
    int signum = 1 ;
    for ( int j = 0 ; j < dimensio ; ++j ) //Abans que res busque el pivot
      {
	float|int max = copia[j][j] ;
	int pivot = j ;
	for ( int i = j + 1 ; i < dimensio ; ++i )
	  {
	    float aij = copia[i][j] ;
	    if ( aij > max )
	      {
		pivot = i ;
		max = aij ;
	      }
	  }
	if ( pivot != j ) 
	  {
	    auxiliar =  NMatrix(copia);
	    auxiliar = auxiliar->swapRows(j, pivot ) ;
	    copia = auxiliar->cast("array");
	    permutacions =  permutacions->swapRows( j,  pivot) ;
	    signum = - signum ;
	  }
     
	// Ja he fet el pivotatge, ara calcule L i U 
	for ( int i = 0 ; i < dimensio ; ++i )
	  {
	    //Rellene la U
	    for ( int j = i ; j < dimensio ; ++j) 
	      {
		float suma = 0 ;
		for (int k = 0 ; k <  j  ; ++k )
		  {
		    suma += l[i][k] * u[k][j] ;
		  }
		u[i][j] = copia[i][j] - suma ;
	      }
	    //Rellene la L
	    for (int j = i + 1 ; j < dimensio ; ++j )
	      {
		float suma = 0.0 ;
		for ( int k = 0; k <  i  ; ++k ) 
		  {
		    suma += l[j][k] *  u[k][i]  ;
		  }
		l[j][i] = ( copia[j][i] - suma ) / (u[i][i] * 1.0 ) ;
	      }
	  }
	return ({NMatrix( l ) , NMatrix ( u ),permutacions, signum } );
      }
  }
  
  //! Given the LU decomposition of a NMatrix, solves the linear equations system AX = w
  public NMatrix backSubstitution( NMatrix L , NMatrix U , NMatrix P ,   NMatrix w) 
  {
    array l = L->cast("array") ;
    array u = U->cast("array") ;
    NMatrix aux =  P->mult(w) ;
    array b = aux->cast("array");
    int dimensio = rows();
    array y = (NMatrix( dimensio , 1 ))->cast("array") ;
    array x = (NMatrix( dimensio , 1  ))->cast("array") ;
    for ( int i = 0 ; i < dimensio ; ++i ) 
      {
	float suma = 0 ;
	for ( int j = 0 ; j <  i  ; ++j )
	  {
	    suma += l[i][j] * y[j][0] ;
	  }
	y[i][0] = ( b[i][0] - suma ) ;
      }
    for ( int i =  ( dimensio - 1 ) ; i > -1 ; --i ) 
      {
	float suma = 0 ;
	for ( int j = i + 1 ; j < dimensio ; ++j )
	  {
	    suma += u[i][j] * x[j][0] ;
	  }
	x[i][0] = ( y[i][0] - suma ) / ( u[i][i] ) ;
      }
    return NMatrix(x) ;
  }

  //!Solves the linear equations system AX = w 
  public NMatrix solveLinearEquations ( NMatrix w ) 
  {
    array ar = LU() ;
    NMatrix l = ar[0] ;
    NMatrix u = ar[1] ;
    NMatrix perm = ar[2] ;
    return  backSubstitution( l, u, perm,   w );
  }

  //!Returns the inverse of the matrix.
  public NMatrix inverse () 
  {
    int dimensio = rows();
    NMatrix aux = NMatrix(dimensio, dimensio );
    array resultat = aux->cast("array");
    NMatrix identitat = NMatrix(dimensio, dimensio);
    array rlu = LU();
    NMatrix l = rlu[0] ;
    NMatrix u = rlu[1] ;
    NMatrix permutacions = rlu[2] ;
    NMatrix columna , solv;
    array mat ;
    
    for ( int i = 0 ; i < dimensio ; ++i )
      {
	columna = identitat->getColumn(i);
	solv = backSubstitution( l, u, permutacions, columna );
	mat = solv->cast("array");
	for ( int j = 0 ; j < dimensio ; ++j )
	  {
	    resultat[j][i] = mat[j][0] ;
	  }
      }
    return NMatrix(resultat);
  }
  //Ull!! Falta comprovar si la matriu és quadrada.
  //!Returns the determinant of a matrix.
  public float determinant () 
  {
    array aux = LU();
    array u = (array) aux[1] ;
    int signum = aux[3] ;
    int dimensio = rows();
    float resultat = 1.0 ;
    for ( int i = 0 ; i < dimensio ; ++i )
      {
	resultat = resultat * u[i][i] ;
      }
    return resultat * signum ;
  }
}

//Funci per a convertir un vector en columna. 
//! Returns a column Matrix from an array
NMatrix ColumnFromArray( array a )
{
  array resultat = (array) NMatrix( sizeof(a), 1) ;
  for ( int i = 0 ; i < sizeof(a) ; ++i )
    {
      resultat[i][0] = a[i] ;
    }
  return NMatrix(resultat);
}

int main () 
{
  NMatrix matriu = NMatrix(2,3);
  NMatrix matriu2 = NMatrix(3, 3) ;
  array sistema = ( { ({1.45 , 0.27 , 3.12 }) , ({ -41.2 , 5.7 ,6.34}) , ({-7.1,8.4,9.3}) } ) ;
  array independent = ( { ({2.45}) , ({1.07}) , ({3.45})   } ) ;
  NMatrix matriu3 = NMatrix( sistema ) ;
  NMatrix w = NMatrix(independent);
  NMatrix l , u, perm ;
  array ar = matriu3->LU();
  l = ar[0] ;
  u = ar[1] ;
  perm = ar[2] ;
  write("The (0,0) element is " + matriu->get(  0 , 0 ) + "\n" );
  write("The dimension of the matrix is " + matriu->dims()[0] + " x " + matriu->dims()[1] + "\n" ) ;
  write("Matrix 2 has dimension " + matriu2->dims()[0] + " x " + matriu2->dims()[1] + "\n"  );
  matriu2->show();
  write("\n");
  write("Transposed matrix\n");
  (matriu->transpose())->show();
  (matriu2->productByScalar(3) )->show();
  write("\n");
  (matriu->swapRows(0,1))->show();
  array resultat = matriu3->LU();;
  write("LU test\n");
  NMatrix PM =  perm * matriu3  ;
  NMatrix  LU =  l *  u ;
  (LU - PM)->show();
  NMatrix resultatSistema =   matriu3->solveLinearEquations(w);
  write("SolveLinearEquations test\n");
  NMatrix MR =   matriu3 * resultatSistema;
  (MR - w)->show();
  NMatrix inversa = matriu3->inverse();
  write("Inverse test\n");
  ( matriu3 * inversa)->show();
  NMatrix determinada = NMatrix( ({ ({ 1.5 , 2.5 }) , ({ 1.0 , 2.0 }) }) ) ;
  float det = determinada->determinant();
  write("Determinant test: " + (det - 0.5) + "\n");

  return 0;
}
