//! @decl array RungeKutta ( float x0 , float xmax , float y0 , float h , function  f )
//! The Runge-Kutta method. Given the point (x0 , y0) / y0 = f(x0) and the step h, solves the differential equation dy / dt = f(x,y) ;
//! @param x0
//! The point where the integration begins
//! @param xmax
//! The point to finish the integration
//! @param y0
//! The initial condition to y.
//! @param f
//! The function which gives df/dt = f(x, y) ;

array RungeKutta ( float x0 , float xmax , float y0 , float h , function  f )
{
  float x = x0, y = y0 , k1, k2 , k3 , k4;
  array resultat = ({ y0 }) ;
  while ( x < xmax  )
    {
      x += h ;
      k1 = h * f(x,y);
      k2 = h * f( x + h/2.0 , y + k1/2.0 );
      k3 = h * f( x + h/2.0 , y + k2/2.0 );
      k4 = h * f( x + h , y + k3 );
      y += 1/6.0 * ( k1 + 2 * k2 + 2 * k3 + k4 ) ;
      resultat += ({ y }) ;
    }
  return resultat ;
}



int main ()
{
  float h = 0.001 ;
  function f = cos ;
  array resultat = RungeKutta( 0.0 , 1.6 , 0.0 , h , f );
  float x = 0.0 ;
  int i = 0 ;
  write("RungeKutta result - Exact result\n");
  foreach( resultat, float item )
    {
      write( "" +  ( item - sin(x)) + "\n");
      x += h ;
    }
  return 0;
}

